# -*- coding: utf-8 -*-
"""
``linea.vector``

Arbitrarily-sized vectors built on numpy's ndarray.

Components:

+ class Vector
+ exception NonConformantVectors
+ function: dot
+ function: angle
+ function: parallel
+ function: orthogonal
"""
# pylint: disable=C0103
import math

import numpy

EQUALITY_TOLERANCE = 0.001


class NonConformantVectors(Exception):
    """
    A NonConformantVector is thrown when attempting to do operations
    with vectors of differing sizes; generally, the size of the
    /right-hand/ argument doesn't conform to the size of the /left-hand/
    argument.
    """

    def __init__(self, expected, actual):
        """
        Initialize a new NonConformantVectors exception.
        :param expected: length of the first vector
        :param actual: length of the second vector
        """
        self.expected = expected
        self.actual = actual
        Exception.__init__(self)

    def __str__(self):
        msg = "Expected the right-hand vector to have dimension {}, "
        msg += "but it has a dimension of {}."
        return msg.format(self.expected, self.actual)


class Vector:
    """
    A vector is a one-dimensional vector of some arbitrary size. This size
    is fixed and can't be changed later in the Vector's life.
    """

    def __init__(self, a=None, *args):
        """
        Initialise a vector, either using an iterable passed in or as a sequence
        of values.
        >>> print(Vector(1, 2, 3))
        [1; 2; 3]
        >>> print(Vector([1, 2, 3]))
        [1; 2; 3]
        """
        if len(args) > 0:
            a = [a]
            a.extend(args)
            self.v = numpy.array(a)
        else:
            if isinstance(a, numpy.ndarray):
                self.v = a
            else:
                self.v = numpy.array(a)

    def __str__(self):
        s = '['
        for i in range(len(self.v)):
            s += str(self.v[i])
            if i < len(self.v) - 1:
                s += '; '
        s += ']'
        return s

    def __len__(self):
        return len(self.v)

    def __iter__(self):
        return (x for x in self.v)

    def __mul__(self, other):
        o = list(map(lambda x: x * other, self.v))
        return Vector(a=o)

    def __rmul__(self, other):
        return self * other

    def __add__(self, other):
        if len(self) != len(other):
            raise NonConformantVectors(len(self), len(other))
        return Vector(self.v + other.v)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        if len(self) != len(other):
            raise NonConformantVectors(len(self), len(other))
        return Vector(self.v - other.v)

    def __eq__(self, other):
        if not isinstance(other, Vector):
            raise ValueError
        if len(self) != len(other):
            raise NonConformantVectors(len(self), len(other))
        eq = numpy.isclose(self.v, other.v, EQUALITY_TOLERANCE)
        if isinstance(eq, bool):
            return eq
        return eq.all()

    def __repr__(self):
        return 'Vector[{}]'.format(len(self))

    def __getitem__(self, item):
        return self.v[item]

    def magnitude(self):
        """
        Return the magnitude of the vector.
        """
        return math.sqrt(sum(map(lambda x: x * x, self.v)))

    def is_zero(self, tolerance=EQUALITY_TOLERANCE):
        """
        Return True if the vector is a zero vector (within some tolerance).
        """
        return numpy.isclose(self.magnitude(), 0, tolerance)

    def unit(self):
        """
        Return the unit vector of this vector. If this method is called on
        a zero vector (i.e. is_zero returns True), a ValueError will be thrown.
        """
        mag = self.magnitude()
        if self.is_zero():
            raise ValueError("cannot normalise the zero vector")
        return self * (1 / mag)

    def dot(self, other):
        """
        Compute the dot product between this vector and the other vector.
        """
        return dot(self, other)

    def angle_with(self, other, in_degrees=False):
        """
        Compute the angle between this vector and the other vector. The
        default is to return the angle in radians. If in_degrees is True,
        returns the angle in degrees.
        """
        return angle(self, other, in_degrees=in_degrees)

    def parallel_to(self, other):
        """Return True if the vector other is parallel to this vector."""
        return parallel(self, other)

    def orthogonal_to(self, other):
        """Return True if the vector other is orthogonal to this vector."""
        return orthogonal(self, other)

    def project_parallel(self, basis):
        """
        Return the projection of this vector onto the given basis vector.
        """
        unit_basis = basis.unit()
        return Vector(dot(self, unit_basis) * unit_basis)

    def project_orthogonal(self, basis):
        """
        Compute the orthogonal projection of the vector from the given basis
        vector.
        """
        spar = self.project_parallel(basis)
        return self - spar


# The dot (or inner) product determines the angle between two vectors.
def dot(v, w):
    """
    Return the dot product of vectors v and w.
    :rtype: int
    """
    assert isinstance(v, Vector)
    assert isinstance(v, Vector)
    inner = sum([a * b for (a, b) in zip(v.v, w.v)])

    # Cauchy-Schwartz inequality
    assert abs(inner) <= (v.magnitude() * w.magnitude())
    return inner


def angle(v, w, in_degrees=False, tolerance=EQUALITY_TOLERANCE):
    """
    Return the angle between vectors v and w in radians. If in_degrees is
    True, return the answer in degrees.
    """
    try:
        # check for floating point problems resulting in domain errors
        inner = dot(v.unit(), w.unit())
        if inner > 1:
            inner = clamp_if_close(inner, 1.0, tolerance)
        if inner < -1:
            inner = clamp_if_close(inner, -1.0, tolerance)
    except NonConformantVectors:
        raise ValueError('Cannot determine the angle between the zero vector and another vector.')
    except:
        # Deliberately pass through any other exceptions.
        raise
    theta = math.acos(inner)
    if in_degrees:
        theta = r2d(theta)
    return theta


def parallel(v, w, tolerance=EQUALITY_TOLERANCE):
    """Return True if vectors v and w are parallel."""
    if len(v) != len(w):
        raise NonConformantVectors(len(v), len(w))
    if v.is_zero() or w.is_zero():
        return True

    # HERE LIES THE RUIN OF A DUMB
    # I'd originally tried to code this by creating a list of v_i / w_i, and then reducing
    # the list via comparison (e.g. are all the values in the list numpy.isclose?). I got it
    # right, but... the video showed a better way. Lesson learned, think about things instead
    # of blindly coding through.
    theta = angle(v, w)
    if numpy.isclose(theta, 0, tolerance):
        return True
    return numpy.isclose(theta, math.pi, tolerance)


def orthogonal(v, w, tolerance=EQUALITY_TOLERANCE):
    """
    Return True if vectors v and w are orthogonal.
    """
    if len(v) != len(w):
        raise NonConformantVectors(len(v), len(w))
    if v.is_zero() or w.is_zero():
        return True
    return numpy.isclose(dot(v, w), 0, tolerance)


def cross(v, w):
    """Return the cross product of the 3D vectors v and w."""
    if len(v) != 3:
        raise NonConformantVectors(3, len(v))
    if len(w) != 3:
        raise NonConformantVectors(3, len(w))

    (x1, y1, z1) = v.v
    (x2, y2, z2) = w.v
    xc = (y1 * z2) - (y2 * z1)
    # - (x1 * z2) - (x2 * z1)
    yc = (x1 * z2) - (x2 * z1)
    # x1 * y2 - x2 * y1
    zc = (x1 * y2) - (x2 * y1)
    return Vector(xc, -yc, zc)


def area_parallelogram(v, w):
    """Return the area of a parallelogram formed by Vectors v and w."""
    u = cross(v, w)
    return u.magnitude()


def area_triangle(v, w):
    """Return the area of a triangle formed by Vectors v and w."""
    return 0.5 * area_parallelogram(v, w)


def r2d(rval):
    """
    Convert the integer or floating point radian value to degrees. The
    radian value must be between :math:`-2*\pi <= rval <= 2*\pi`.

    >>> r2d(math.pi)
    180.0
    >>> r2d(3 * math.pi / 4)
    135.0
    >>> r2d(3 * math.pi)
    ---------------------------------------------------------------------------
    AssertionError                            Traceback (most recent call last)
    """
    assert abs(rval) <= (2 * math.pi)
    return rval * 180 / math.pi


def clamp_if_close(value, clamped, tolerance=EQUALITY_TOLERANCE):
    """
    Return clamped if value is close to clamped (within tolerance), or return
    value otherwise.

    >>> clamp_if_close(0.99, 1.0, tolerance=0.1)
    1.0
    >>> clamp_if_close(0.99, 1.0, tolerance=0.001)
    0.99
    """
    if numpy.isclose(value, clamped, tolerance):
        return clamped
    return value
