import math

import numpy


class Rational:
    """
    Rational implements a rational number type.
    """

    def __init__(self, n, d):
        self.n = int(n)
        self.d = int(d)

        # hack to avoid DBZ
        if d == 0:
            self.n = 1
            self.d = 0

    def __float__(self) -> float:
        return self.n / self.d

    def __int__(self) -> int:
        """
        The integer version of a Rational is the rounded float representation
        of the number.
        """
        return round(self.n / self.d)

    def __add__(self, other):
        lcm = int(math.lcm(self.d, other.d))
        ks = lcm / self.d
        ko = lcm / other.d

        return Rational((self.n * ks) + (other.n * ko), lcm)

    def __mul__(self, other):
        return Rational(self.n * other.n, self.d * other.d)

    def __truediv__(self, other):
        return Rational(self.n * other.d, self.d * other.n)

    def lcm(self, other):
        return int(numpy.lcm(self.d, other.d))

    def normalize(self, other):
        lcm = self.lcm(other)
        k = lcm / self.d
        return self.scale(k)

    def __normpair__(self, other):
        rs = self.reduce()
        ro = other.reduce()
        ns = rs.normalize(ro)
        no = ro.normalize(rs)
        return ns, no

    def reduce(self):
        gcd = math.gcd(self.n, self.d)
        return Rational(self.n / gcd, self.d / gcd)

    def __eq__(self, other):
        rs = self.reduce()
        ro = other.reduce()
        return (rs.n == ro.n) and (rs.d == ro.d)

    def __gt__(self, other):
        ns, no = self.__normpair__(other)
        return ns.n > no.n

    def __ge__(self, other):
        ns, no = self.__normpair__(other)
        return ns.n >= no.n

    def __lt__(self, other):
        ns, no = self.__normpair__(other)
        return ns.n < no.n

    def __le__(self, other):
        ns, no = self.__normpair__(other)
        return ns.n <= no.n

    def scale(self, k):
        return Rational(self.n * k, self.d * k)

    def __sub__(self, other):
        lcm = self.lcm(other)
        ks = lcm / self.d
        ko = lcm / other.d

        return Rational((self.n * ks) - (other.n * ko), lcm)

    def __pow__(self, power):
        return Rational(self.n ** power, self.d ** power)

    def __repr__(self):
        return f'Rational({self.n}, {self.d})'


def Int(n) -> Rational:
    """
    Int produces an integer version of a Rational (e.g. denominator is 1).
    """
    return Rational(n, 1)


def ipow(n, m):
    x = n

    while m > 1:
        x = x * n
        m -= 1

    return x


def expansion(v):
    top = Int(1)
    bottom = Int(2) + v

    return top / bottom


def recurse(steps, start=None):
    x = start
    if x is None:
        x = Rational(1, 2)
    print(f'{x}')

    while steps > 0:
        x = expansion(x)
        print(f'{x}')
        steps -= 1

    return x + Int(1)
